#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
from sysrepo_api import sysrepo_api3 as api
from oslo_utils import netutils


def _isIPv6(ipaddr):
    """Check the IP address version and return True in case of IPv6"""
    if ipaddr is not None and ipaddr != "":
        return netutils.is_valid_ipv6(ipaddr)

    return False


def _prepareforUrlIPv6(ipaddr):
    """Returns ip address suitable for URLs (add square brackets to IPv6 addresses)"""
    # Check is IP is v6
    if _isIPv6(ipaddr):
        # Only if both square brackets are not present, only then they will be added
        if '[' not in ipaddr and ']' not in ipaddr:
            return '[{}]'.format(ipaddr)
    # No preparation needed
    return ipaddr

class SysrepoSNMP(api.SysrepoAPI):

    def __init__(self,
                 app_name="snmp_server",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/snmp-server:snmp-server",
                 indexes={"views": "id", "users": "id", "target-address": "id"},
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority)

    def _merge_changes(self, change_set):
        cfg = self.get_running_config(full_path=False)
        if change_set:
            for change in change_set:
                xpath = change['xpath']
                if not change['xpath'].startswith(self.namespace):
                    continue
                for list_name in ['views', 'users', 'target-address']:
                    if xpath.find(list_name) < 0:
                        continue

                    list_data = []
                    if list_name in cfg:
                        list_data = cfg[list_name]

                    key = str(xpath)[xpath.find("[id='") + 5:xpath.find("']")]
                    field = str(xpath)[xpath.find("']/") + len("']/"):]
                    for item in list_data:
                        if 'id' in item and item['id'] == int(key):
                            if change['operation'] == 'DELETED':
                                if field in item:
                                    del item[field]
                            else:
                                item[field] = change['new_value']

                            break

                    break

                if change['operation'] == 'DELETED':
                    continue

                for el in ['enable', 'bind-address', 'port', 'authentication-failure-trap-enabled', 'system-name',
                           'system-location', 'system-contact', 'system-description']:
                    if xpath.find(el) < 0:
                        continue

                    cfg[el] = change['new_value']
                    break

        return cfg

    def apply_config(self, change_set=None, apply_startup=False):
        if apply_startup:
            cfg = self.get_running_config(full_path=False)
        else:
            cfg = self._merge_changes(change_set)

        with open('/etc/snmp/snmpd.conf', 'w') as config_file:
            config_file.write('### AgentX Sub-Agents\n')
            config_file.write('master agentx\n')
            #
            config_file.write('\n### System information\n')
            if 'system-name' in cfg and cfg['system-name'] != u'':
                config_file.write('sysname ' + cfg['system-name'] + '\n')

            if 'system-location' in cfg and cfg['system-location'] != u'':
                config_file.write('syslocation ' + cfg['system-location'] + '\n')

            if 'system-contact' in cfg and cfg['system-contact'] != u'':
                config_file.write('syscontact ' + cfg['system-contact'] + '\n')

            if 'system-description' in cfg and cfg['system-description'] != u'':
                config_file.write('sysdescr ' + cfg['system-description'] + '\n')

            #
            config_file.write('sysServices 11\n')
            #
            config_file.write('\n### Agent Behaviour\n')
            if 'bind-address' in cfg:
                if _isIPv6(cfg['bind-address']) is True:
                #IPv6
                    config_file.write('agentaddress udp6:' + _prepareforUrlIPv6(cfg['bind-address']) + ':' +\
                                      str(cfg['port']) + '\n')
                else:
                # IPv4
                    config_file.write('agentaddress udp:' + cfg['bind-address'] + ':' +\
                                      str(cfg['port']) + '\n')
            #
            config_file.write('\n### VACM Configuration\n')
            config_file.write('\n# View\n')
            if 'views' in cfg:
                for views in cfg['views']:
                    config_file.write('view ' + views['name'] + ' ' + views['type'] + ' ' + views['oid'] + '\n')
                    self.execute_shell_command("systemctl stop snmpd.service")

            #
            config_file.write('\n# Access Control\n')
            if 'users' in cfg:
                for users in cfg['users']:
                    if users['message-model'] == 'v1' or users['message-model'] == 'v2c':
                        config_file.write('com2sec ' + \
                                          users['user-name'] + ' default ' + \
                                          users['user-name'] + '\n')
                        config_file.write('com2sec6 ' + \
                                          users['user-name'] + ' default ' + \
                                          users['user-name'] + '\n')
                        config_file.write(
                            'group ' + users['user-name'] + "Group " + \
                            users['message-model'] + ' ' + \
                            users['user-name'] + '\n')
                        access = 'access ' + users['user-name'] + "Group \"\" " + \
                                 users['message-model'] + ' ' + \
                                 users['security-level'] + ' exact '

                        if 'read-view' in users and users['read-view'] != "":
                            access = access + users['read-view'] + ' '
                        else:
                            access = access + "none "

                        if 'write-view' in users and users['write-view'] != "":
                            access = access + users['write-view'] + ' '
                        else:
                            access = access + "none "

                        if 'notify-view' in users and users['notify-view'] != "":
                            access = access + users['notify-view'] + '\n'
                        else:
                            access = access + "none\n"

                        #
                        config_file.write(access)

            #
            config_file.write('rouser internal\n')
            #
            config_file.write('\n### Active Monitoring\n')
            config_file.write('\n# Notification Handling\n')
            config_file.write('trapcommunity monitoring\n')
            #
            if 'target-address' in cfg:
                for targets in cfg['target-address']:
                    if 11 != len(targets.keys()):
                        # all keys must exists - they are:
                        # id,enable,dst-port,dst-address,message-model,
                        # retry-count,security-level,security-name,
                        # target-addr-name,timeout,type
                        continue

                    for targets_values in targets.values():
                        if '' == str(targets_values):
                            # no empty values
                            continue

                    if 'false' == str(targets['enable']).lower():
                        trapsess = '# trapsess -'
                    else:
                        trapsess = 'trapsess -'

                    trapsess = trapsess + targets['message-model']
                    if targets['message-model'] == 'v3':
                        for targetuser in cfg['users']:
                            if targets['security-name'] == targetuser['user-name']:
                                trapsess = trapsess + ' -u ' + \
                                           targets['security-name'] + ' -l ' + \
                                           targets['security-level'] + ' -a ' + \
                                           targetuser['auth-type'] + ' -A ' + \
                                           targetuser.get('auth-password','\"\"') + ' -x ' + \
                                           targetuser['priv-type'] + ' -X ' + \
                                           targetuser.get('priv-password','\"\"')
                    else:
                        trapsess = trapsess + ' -c ' + targets['security-name']

                    if targets['type'] == 'inform':
                        trapsess = trapsess + ' -Ci ' + \
                                   targets['dst-address'] + ':' + \
                                   str(targets['dst-port']) + ' -r ' + \
                                   str(targets['retry-count']) + ' -t ' + \
                                   str(targets['timeout']) + '\n'
                    else:
                        trapsess = trapsess + ' ' + \
                                   targets['dst-address'] + ':' + \
                                   str(targets['dst-port']) + ' -r ' + \
                                   str(targets['retry-count']) + ' -t ' + \
                                   str(targets['timeout']) + '\n'

                    #
                    config_file.write(trapsess)
                    #
                    if 1 == targets['id']:
                        # first target goes to delayed_reboot
                        new_line_in_trapsses = len(trapsess)
                        if 'false' == str(targets['enable']).lower():
                            new_reboot_target = "# /usr/bin/snmptrap" + trapsess[10:new_line_in_trapsses - 1] + \
                                                " SNMPv2-MIB::warmStart 2>/dev/null \n"
                        else:
                            new_reboot_target = "/usr/bin/snmptrap" + trapsess[8:new_line_in_trapsses - 1] + \
                                                " SNMPv2-MIB::warmStart 2>/dev/null \n"

                        reboot_file_lines = str()
                        with open('/scripts/delayed_reboot.sh', 'r') as reboot_file:
                            reboot_file_lines = reboot_file.readlines()
                            reboot_file.close()

                        snmptrap_indent_position = 0
                        new_reboot_file_content = []
                        for reboot_line in reboot_file_lines:
                            snmptrap_indent_position = reboot_line.find('/usr/bin/snmptrap')
                            if 0 <= snmptrap_indent_position:
                                reboot_line = reboot_line[0:2] + new_reboot_target

                            new_reboot_file_content.append(reboot_line)

                        with open('/scripts/delayed_reboot.sh', 'w') as reboot_file:
                            reboot_file.writelines(new_reboot_file_content)
                            reboot_file.flush()
                            os.fsync(reboot_file.fileno())
                            reboot_file.close()

            #
            config_file.write('# DisMan Event MIB\n')
            config_file.write('agentSecName internal\n')
            config_file.write('notificationEvent  linkUpTrap    IF-MIB::linkUp   ifIndex ifAdminStatus ifOperStatus\n')
            config_file.write('notificationEvent  linkDownTrap  IF-MIB::linkDown ifIndex ifAdminStatus ifOperStatus\n')
            config_file.write('monitor -r 1 -e linkUpTrap "Generate linkUp" ifOperStatus != 2\n')
            config_file.write('monitor -r 1 -e linkDownTrap "Generate linkDown" ifOperStatus == 2\n')
            config_file.write('sysObjectID .1.3.6.1.4.1.738.20.50001\n')
            config_file.write('# Send traps on authentication failures\n')
            if 'authentication-failure-trap-enabled' in cfg and cfg['authentication-failure-trap-enabled'] is True:
                config_file.write('authtrapenable 1\n')
            else:
                config_file.write('authtrapenable 2\n')

            #
            config_file.write('maxGetbulkResponses 20\n')

            config_file.close()

        v3_users = False
        if 'users' in cfg:
            for users in cfg['users']:
                if users['message-model'] == 'v3':
                    v3_users = True
                    break

        # Set v3
        if v3_users:
            if 'users' in cfg:
                for users in cfg['users']:
                    if users['message-model'] == 'v3':
                        v3_user_command = "net-snmp-create-v3-user" + \
                                          " -A " + users.get('auth-password','\"\"') + \
                                          " -X " + users.get('priv-password','\"\"') + \
                                          " -a " + users['auth-type'] + \
                                          " -x " + users['priv-type'] + \
                                          " " + users['user-name']
                        self.execute_shell_command(v3_user_command)

        if 'enable' in cfg and 'true' == str(cfg['enable']).lower():
            self.execute_shell_command("systemctl enable snmpd.service")
            if True == self.is_service_active('snmpd.service'):
                snmpd_pid = str(self.execute_shell_command("pidof snmpd")).strip("\n")
                self.execute_shell_command("kill -HUP {}".format(snmpd_pid))
            else:
                self.execute_shell_command("systemctl --no-block start snmpd.service")

        else:
            self.execute_shell_command("systemctl disable snmpd.service")
            self.execute_shell_command("systemctl stop snmpd.service")

        return True

    def set_enable(self, enable=False):
        self.set_item("{}/enable".format(self.namespace), enable)

    def get_enable(self):
        return self.get_item("{}/enable".format(self.namespace))

    def unset_enable(self):
        self.delete_item("{}/enable".format(self.namespace))

    def set_bind_address(self, bind_address):
        self.set_item("{}/bind-address".format(self.namespace), bind_address)

    def get_bind_address(self):
        return self.get_item("{}/bind-address".format(self.namespace))

    def unset_bind_address(self):
        self.delete_item("{}/bind-address".format(self.namespace))

    def set_port(self, port=161):
        self.set_item("{}/port".format(self.namespace), port)

    def get_port(self):
        return self.get_item("{}/port".format(self.namespace))

    def unset_port(self):
        self.delete_item("{}/port".format(self.namespace))

    def set_authentication_failure_trap_enabled(self, authentication_failure_trap_enabled=False):
        self.set_item("{}/authentication-failure-trap-enabled".format(self.namespace),
                      authentication_failure_trap_enabled)

    def get_authentication_failure_trap_enabled(self):
        return self.get_item("{}/authentication-failure-trap-enabled".format(self.namespace))

    def unset_authentication_failure_trap_enabled(self):
        self.delete_item("{}/authentication-failure-trap-enabled".format(self.namespace))

    def set_system_name(self, system_name):
        self.set_item("{}/system-name".format(self.namespace), system_name)

    def get_system_name(self):
        return self.get_item("{}/system-name".format(self.namespace))

    def unset_system_name(self):
        self.delete_item("{}/system-name".format(self.namespace))

    def set_system_location(self, system_location):
        self.set_item("{}/system-location".format(self.namespace), system_location)

    def get_system_location(self):
        return self.get_item("{}/system-location".format(self.namespace))

    def unset_system_location(self):
        self.delete_item("{}/system-location".format(self.namespace))

    def set_system_contact(self, system_contact):
        self.set_item("{}/system-contact".format(self.namespace), system_contact)

    def get_system_contact(self):
        return self.get_item("{}/system-contact".format(self.namespace))

    def unset_system_contact(self):
        self.delete_item("{}/system-contact".format(self.namespace))

    def set_system_description(self, system_description):
        self.set_item("{}/system-description".format(self.namespace), system_description)

    def get_system_description(self):
        return self.get_item("{}/system-description".format(self.namespace))

    def unset_system_description(self):
        self.delete_item("{}/system-description".format(self.namespace))

    def set_view(self, id):
        self.set_item("{}/views[id='{}']".format(self.namespace, id), id)

    def get_view(self, id):
        return self.get_item("{}/views[id='{}']".format(self.namespace, id))

    def unset_view(self, id):
        self.delete_item("{}/views[id='{}']".format(self.namespace, id))

    def get_views(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'views' not in config:
            return config_indexed

        for view in config['views']:
            config_indexed[view['id']] = view

        return config_indexed

    def set_view_param(self, id, param, value):
        self.set_item("{}/views[id='{}']/{}".format(self.namespace, id, param), value)

    def get_view_param(self, id, param):
        return self.get_item("{}/views[id='{}']/{}".format(self.namespace, id, param))

    def unset_view_param(self, id, param):
        self.delete_item("{}/views[id='{}']/{}".format(self.namespace, id, param))

    def set_user(self, id):
        self.set_item("{}/users[id='{}']".format(self.namespace, id), id)

    def get_user(self, id):
        return self.get_item("{}/users[id='{}']".format(self.namespace, id))

    def unset_user(self, id):
        self.delete_item("{}/users[id='{}']".format(self.namespace, id))

    def get_users(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'users' not in config:
            return config_indexed
        for user in config['users']:
            config_indexed[user['id']] = user
        return config_indexed

    def set_user_param(self, id, param, value):
        self.set_item("{}/users[id='{}']/{}".format(self.namespace, id, param), value)

    def get_user_param(self, id, param):
        return self.get_item("{}/users[id='{}']/{}".format(self.namespace, id, param))

    def unset_user_param(self, id, param):
        self.delete_item("{}/users[id='{}']/{}".format(self.namespace, id, param))

    def set_target_address(self, id):
        self.set_item("{}/target-address[id='{}']".format(self.namespace, id), id)

    def get_target_address(self, id):
        return self.get_item("{}/target-address[id='{}']".format(self.namespace, id))

    def get_target_addresses(self):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'target-address' not in config:
            return config_indexed
        for target_address in config['target-address']:
            config_indexed[target_address['id']] = target_address
        return config_indexed

    def unset_target_address(self, id):
        self.delete_item("{}/target-address[id='{}']".format(self.namespace, id))

    def set_target_address_param(self, id, param, value):
        self.set_item("{}/target-address[id='{}']/{}".format(self.namespace, id, param), value)

    def get_target_address_param(self, id, param):
        return self.get_item("{}/target-address[id='{}']/{}".format(self.namespace, id, param))

    def unset_target_address_param(self, id, param):
        self.delete_item("{}/target-address[id='{}']/{}".format(self.namespace, id, param))
